from .common import vcr
from asnake.aspace import ASpace
from asnake.jsonmodel import JSONModelObject, TreeNode, ComponentObject, JSONModelRelation, searchdoc_signifiers
import os

conf_file = None
aspace = None

def setup():
    global aspace, conf_file
    '''Point ASNAKE_CONFIG_FILE at non-extant path so local config DEF HAPPENS even if you have a config'''
    try:
        conf_file = os.environ.pop('ASNAKE_CONFIG_FILE')
    except: pass
    os.environ['ASNAKE_CONFIG_FILE'] = "NONSENSEFILETHATDOESNOTEXIST"
    aspace = ASpace()

@vcr.use_cassette
def test_fetch():
    assert isinstance(aspace.repositories, JSONModelRelation)
    resolved = list(aspace.repositories)
    assert resolved[0].jsonmodel_type == "repository"
    repo_id = resolved[0].uri.split("/")[-1]
    assert isinstance(aspace.repositories(repo_id), JSONModelObject)

@vcr.use_cassette
def test_ordered_records():
    getem = list(aspace.resources)[0].ordered_records.uris
    assert len(getem) > 0
    assert all(isinstance(x, JSONModelObject) for x in getem)

@vcr.use_cassette
def test_stays_ref_on_repr():
    agent = aspace.repositories(2).agent_representation
    assert(agent.is_ref)
    agent.__repr__()
    assert(agent.is_ref)

@vcr.use_cassette
def test_agent_subroute():
    corp_ent = aspace.agents.corporate_entities
    assert isinstance(corp_ent, JSONModelRelation)
    # raised prior due to subroutes getting AgentRelation type
    agent = aspace.agents.corporate_entities(1)

@vcr.use_cassette
def test_by_external_id():
    result = next(aspace.by_external_id('thingumie'))
    assert isinstance(result, JSONModelObject)

@vcr.use_cassette
def test_trees():
    resource_tree = aspace.repositories(2).resources(1).tree
    assert isinstance(resource_tree.children[0], TreeNode)
    records_via_walk = list(resource_tree.walk)
    assert aspace.repositories(2).resources(1).json() == records_via_walk[0].json()
    assert isinstance(records_via_walk[1], ComponentObject)
    subtree_walk = list(records_via_walk[1].tree.walk)
    assert records_via_walk[1].uri == subtree_walk[0].uri

@vcr.use_cassette
def test_with_params():
    list(aspace.repositories(2).search.with_params(q="primary_type:resource AND publish:true"))

@vcr.use_cassette
def test_solr_route():
    list(aspace.repositories(2).top_containers.search.with_params(q="barcode_field:1234"))

@vcr.use_cassette
def test_from_uri():
    repo = aspace.from_uri('/repositories/2')
    assert isinstance(repo, JSONModelObject)
    assert repo.uri == '/repositories/2'
    repo.name # just making sure a known present method works

@vcr.use_cassette
def test_users_route():
    for user in  aspace.users:
        assert isinstance(user.permissions, dict)

@vcr.use_cassette
def test_search_route_unwrapping():
    # Get first ao, just really care that it ends up Not Wrapped
    ao = next(iter(aspace.repositories(2).search.with_params(q="*:* and primary_type:archival_object")))
    assert set(ao.json().keys()).isdisjoint(searchdoc_signifiers)

def teardown():
    '''Undo the thing from setup'''
    if conf_file:
        os.environ['ASNAKE_CONFIG_FILE'] = conf_file
    else:
        os.environ.pop('ASNAKE_CONFIG_FILE')



from asnake import ASnakeConfig
import os
from os.path import dirname, join

def setup():
    os.environ['ASNAKE_CONFIG_FILE'] = join(dirname(__file__), 'fixtures', 'asnake.yml')

def test_config_from_file():
    expected = {
        "username": "APIUser",
        "password": "APIPass",
        "baseurl":  "http://notarealurl.com:4567"
    }

    config = ASnakeConfig()
    for k in expected:
        assert expected[k] == config[k]


def test_direct_config():
    expected = {
        "username": "directUser",
        "password": "directPass",
        "baseurl": "directNotRealUrl"
    }

    config = ASnakeConfig()
    config.update(expected)
    for k in expected:
        assert expected[k] == config[k]


def teardown():
    os.environ.pop('ASNAKE_CONFIG_FILE')



from requests import Session
from urllib.parse import quote
from numbers import Number
from collections.abc import Sequence, Mapping

import json
import asnake.configurator as conf
import asnake.logging as logging

log = None # initialized on first client init

class ASnakeAuthError(Exception): pass
class ASnakeWeirdReturnError(Exception): pass

def listlike_seq(seq):
    '''Determine if a thing is a list-like (sequence of values) sequence that's not string-like.'''
    return isinstance(seq, Sequence) and not isinstance(seq, (str, bytes, Mapping,))

def http_meth_factory(meth):
    '''Utility method for producing HTTP proxy methods for ASnakeProxyMethods mixin class.

    Urls are prefixed with the value of baseurl from the client's ASnakeConfig.  Arguments are
    passed unaltered to the matching requests.Session method.'''
    def http_method(self, url, *args, **kwargs):
        # aspace uses the PHP convention where array-typed form values use names with '[]' appended
        if 'params' in kwargs:
            kwargs['params'] = {k + '[]' if listlike_seq(v) and k[-2:] != '[]' else k:v for k,v in kwargs['params'].items()}

        full_url = "/".join([self.config['baseurl'].rstrip("/"), url.lstrip("/")])
        result = getattr(self.session, meth)(full_url, *args, **kwargs)
        if result.status_code == 403 and self.config['retry_with_auth']:
            self.authorize()
            result = getattr(self.session, meth)(full_url, *args, **kwargs)
        log.debug("proxied http method", method=meth.upper(), url=full_url, status=result.status_code)
        return result
    return http_method

class ASnakeProxyMethods(type):
    '''Metaclass to set up proxy methods for all requests-supported HTTP methods'''
    def __init__(cls, name, parents, dct):

        for meth in ('get', 'post', 'head', 'put', 'delete', 'options',):
            fn = http_meth_factory(meth)
            fn.__name__ = meth
            fn.__doc__ = '''Proxied :meth:`requests.Session.{}` method from :class:`requests.Session`'''.format(meth)

            setattr(cls, meth, fn)

class ASnakeClient(metaclass=ASnakeProxyMethods):
    '''ArchivesSnake Web Client'''

    def __init__(self, **config):
        global log

        if 'config_file' in config:
            self.config = conf.ASnakeConfig(config['config_file'])
        else:
            self.config = conf.ASnakeConfig()

        self.config.update(config)

        # Only a subset of logging config can be supported in config
        # For more complex setups (configuring output format, say),
        # configure logs in Python code prior to loading
        #
        # Properties supported are:
        #    filename, filemode, level, and default_config
        # Default config can be any of the default configurations exposed in logging
        if not log:
            if not logging.already_configured and 'logging_config' in self.config:
                if 'default_config' in self.config['logging_config']:
                    default_logging_config = logging.configurations.get(
                        self.config['logging_config']['default_config'])
                    del self.config['logging_config']['default_config']
                else:
                    default_logging_config = None

                logging.setup_logging(config = default_logging_config,
                                      **self.config['logging_config'])

            log = logging.get_logger(__name__)

        if not hasattr(self, 'session'): self.session = Session()
        self.session.headers.update({'Accept': 'application/json',
                                     'User-Agent': 'ArchivesSnake/0.1'})
        log.debug("client created")


    def authorize(self, username=None, password=None):
        '''Authorizes the client against the configured archivesspace instance.

        Parses the JSON response, and stores the returned session token in the session.headers for future requests.
        Asks for a "non-expiring" session, which isn't truly immortal, just long-lived.'''

        username = username or self.config['username']
        password = password or self.config['password']

        log.debug("authorizing against ArchivesSpace", user=username)

        resp = self.session.post(
            "/".join([self.config['baseurl'].rstrip("/"), 'users/{username}/login']).format(username=quote(username)),
            params={"password": password, "expiring": False}
        )

        if resp.status_code != 200:
            log.debug("authorization failure", status=resp.status_code)
            raise ASnakeAuthError("Failed to authorize ASnake with status: {}".format(resp.status_code))
        else:
            session_token = json.loads(resp.text)['session']
            self.session.headers['X-ArchivesSpace-Session'] = session_token
            log.debug("authorization success", session_token=session_token)
            return session_token


    def get_paged(self, url, *args, page_size=100, **kwargs):
        '''get list of json objects from urls of paged items'''
        params = {}

        if "params" in kwargs:
            params.update(**kwargs['params'])
            del kwargs['params']

        # special-cased bc all_ids doesn't work on repositories index route
        if "all_ids" in params and url in {"/repositories", "repositories"}:
            del params['all_ids']

        params.update(page_size=page_size, page=1)

        current_page = self.get(url, params=params, **kwargs)
        current_json = current_page.json()
        # Regular paged object
        if hasattr(current_json, 'keys') and \
           {'results', 'this_page', 'last_page'} <= set(current_json.keys()):
            while current_json['this_page'] <= current_json['last_page']:
                for obj in current_json['results']:
                    yield obj
                if current_json['this_page'] == current_json['last_page']: break
                params['page'] += 1
                current_page = self.get(url, params=params)
                current_json = current_page.json()
        # routes that just return a list,  or ids, i.e. queries with all_ids param
        elif isinstance(current_json, list):
            # e.g. repositories
            if len(current_json) >= 1:
                if hasattr(current_json[0], 'keys'):
                    for obj in current_json:
                        yield obj
                elif isinstance(current_json[0], Number):
                    for i in current_json:
                        yield self.get("/".join([url, str(i)])).json()
                else:
                    raise ASnakeWeirdReturnError("get_paged doesn't know how to handle {}".format(current_json))
        else:
            raise ASnakeWeirdReturnError("get_paged doesn't know how to handle {}".format(current_json))



import logging
import sys, os
import structlog
from copy import copy
import re

def copy_config(config):
    '''Copy relevant information from one config to another.'''

    new_config = {}
    new_logging = config['logging'].copy()
    new_structlog = {k:copy(v) for k,v in config['structlog'].items()}

    new_config.update(logging=new_logging, structlog=new_structlog)
    return new_config

# Regex to Match down or mixed-case usage of standard logging levels
canonical_levels = ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET',)
level_re = re.compile(
    r"|".join((r'^{}$'.format(level) for level in canonical_levels)),
    re.I
)

handler = None
already_configured = False
def setup_logging(config=None, level=None, stream=None, filename=None, filemode=None):
    '''sets up both logging and structlog.'''
    global handler, already_configured

    asnake_root_logger = logging.getLogger('asnake')

    if handler:
        asnake_root_logger.removeHandler(handler)

    if stream and filename:
        raise RuntimeError("stream and filename are mutually exclusive and cannot be combined, pick one or the other")
    from_env = os.environ.get('ASNAKE_LOG_CONFIG', None)
    default = configurations.get(from_env, DEFAULT_CONFIG)

    if not config:
        config = copy_config(default)
        if filename:
            del config['logging']['stream']

    level = level or config.get('level', None) or logging.INFO
    if isinstance(level, str) and level_re.match(level):
        level = getattr(logging, level.upper())

    # Forward what's needed to put the log places
    if stream:
        config['logging']['stream'] = stream
    if filemode:
        config['logging']['filemode'] = filemode
    if filename:
        config['logging']['filename'] = filename


    if 'filename' in config['logging']:
        handler = logging.FileHandler(config['logging']['filename'],
                                      mode=config['logging'].get('filemode', 'a'))
    if 'stream' in config['logging']:
        handler = logging.StreamHandler(config['logging']['stream'])


    asnake_root_logger.addHandler(handler)
    asnake_root_logger.setLevel(level)
    structlog.reset_defaults()
    structlog.configure(**config['structlog'])
    already_configured = True

def get_logger(name=None):
    if not already_configured:
        setup_logging()

    # Make sure it's under the root logger
    if name and not name.startswith('asnake.'):
        name = 'asnake.' + name

    return structlog.get_logger(name)

# Log format is standard across all provided defaults
# This amounts to a log of serialized JSON events with UTC timestamps and various
# useful information attached, which formats exceptions passed to logging methods in exc_info
def default_structlog_conf(**overrides):
    '''Generate a default configuration for structlog'''
    conf = {
        "logger_factory": structlog.stdlib.LoggerFactory(),
        "wrapper_class":structlog.stdlib.BoundLogger,
        "cache_logger_on_first_use": True,
        "processors": [
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.TimeStamper(fmt='iso', utc=True),
            structlog.processors.JSONRenderer()]
    }
    conf.update(**overrides)
    return conf

def default_logging_conf(**overrides):
    '''Generate a default stdlib logging configuration.'''
    conf = {"level": logging.INFO, "format": "%(message)s"}
    conf.update(**overrides)
    return conf

# Provided example configurations for structlog.
configurations = {}

configurations['INFO_TO_STDERR'] = {
    "logging": default_logging_conf(stream=sys.stderr),
    "structlog": default_structlog_conf()
}

configurations['INFO_TO_STDOUT'] = {
    "logging": default_logging_conf(stream=sys.stdout),
    "structlog": default_structlog_conf()
}

configurations['INFO_TO_FILE'] = {
    "logging": default_logging_conf(level=logging.INFO,
                                    filename=os.path.expanduser("~/archivessnake.log"),
                                    filemode="a"),
    "structlog": default_structlog_conf(),
    'level': 'INFO',
}

configurations['DEBUG_TO_STDERR'] = {
    "logging": default_logging_conf(level=logging.DEBUG, stream=sys.stderr),
    "structlog": default_structlog_conf(),
    "level": "DEBUG"
}

configurations['DEBUG_TO_STDOUT'] = {
    "logging": default_logging_conf(level=logging.DEBUG, stream=sys.stdout),
    "structlog": default_structlog_conf(),
    "level": "DEBUG"
}

configurations['DEBUG_TO_FILE'] = {
    "logging": default_logging_conf(level=logging.DEBUG,
                                    filename=os.path.expanduser("~/archivessnake.log"),
                                    filemode="a"),
    "structlog": default_structlog_conf(),
    'level': 'DEBUG',
}

configurations['DEFAULT_CONFIG'] = configurations['INFO_TO_STDERR']

# Expose these as constants for user convenience
globals().update(**configurations)



from .common import vcr
from asnake.client import ASnakeClient
import os

conf_file = None

def setup():
    '''Point ASNAKE_CONFIG_FILE at non-extant path so local config DEF HAPPENS even if you have a config'''
    try:
        conf_file = os.environ.pop('ASNAKE_CONFIG_FILE')
    except: pass
    os.environ['ASNAKE_CONFIG_FILE'] = "NONSENSEFILETHATDOESNOTEXIST"

@vcr.use_cassette
def test_authorize():
    client = ASnakeClient() # relies on default config, see ASnakeConfig class
    toke = client.authorize()
    assert isinstance(toke, str)
    assert len(toke) == 64
    assert set(toke) <= set('0123456789abcdef')
    assert client.session.headers['X-ArchivesSpace-Session'] == toke
    # Try to get admin user info, should only work if we're authed as admin
    assert client.get('users/1').status_code == 200

def teardown():
    '''Undo the thing from setup'''
    if conf_file:
        os.environ['ASNAKE_CONFIG_FILE'] = conf_file
    else:
        os.environ.pop('ASNAKE_CONFIG_FILE')
